% [MM,LL] = kmodes(X,K[,s,I,J,tol,C]) K-modes clustering of dataset X
%
% In:
%   X: NxD matrix containing N D-dimensional data points rowwise.
%   K: integer in [1,N] containing the desired number of clusters.
%   s: if a scalar, the bandwidth for the kernel density estimate; if a list
%      of scalars, the bandwidths to be used sequentially in the homotopy.
%      Default: calculated (this gives a range between a large bandwidth
%      (k-means) and a target bandwidth that gives a good density estimate).
%   I: maximal number of mean-shift iterations in the mode-finding step.
%      Default: 100.
%   J: maximal number of k-modes iterations (for each bandwidth). Default: 100.
%   tol: small positive number, tolerance in the relative change of the
%      centroids to stop iterating. Default: 1e-3.
%   C: KxD matrix containing the initial K centroids. Default: from k-means.
% Out:
%   MM,LL: cell arrays where each element corresponds to each bandwidth
%      value in s:
%      MM: list of KxD matrices containing the K cluster modes.
%      LL: list of Nx1 vectors containing the cluster labels (1 to K).
%
% Any non-mandatory argument can be given the value [] to force it to take
% its default value.

% Copyright (c) 2013 by Miguel A. Carreira-Perpinan and Weiran Wang

function [MM,LL] = kmodes(X,K,s,I,J,tol,C)

% ---------- Argument defaults ----------
if ~exist('s','var') || isempty(s) s = bandwidths(X); end;
if ~exist('I','var') || isempty(I) I = 100; end;
if ~exist('J','var') || isempty(J) J = 100; end;
if ~exist('tol','var') || isempty(tol) tol = 1e-3; end;
if ~exist('C','var') || isempty(C) C = 'kmeans'; end;
% ---------- End of "argument defaults" ----------

s = unique(s); s = s(end:-1:1);
ls = length(s); MM = cell(1,ls+1); LL = cell(1,ls+1);

% Initial assignments
M = km_init(X,K,C); l = km_le(X,M); MM{1} = M; LL{1} = l;

% Homotopy path, for every value of the bandwidth
for i=1:ls
  sigma = s(i);
  if sigma>0
    % Run at most J iterations of K-modes for fixed bandwidth = sigma
    for j=1:J
      oldM = M;
      M = km_m(X,sigma,l,M,I,tol);	% Find modes given assignments
      l = km_le(X,M);			% Find assignments given modes
      if max(abs(M(:)-oldM(:)) - tol*abs(oldM(:))) < 0 break; end
    end
  else
    % When the bandwidth=0, the kde is a needle plot, so we approximate the
    % mode finding in each cluster by nearest-neighbor finding.
    for k=1:K
      tmp = find(l==k); [~,idx] = min(sqdist(M(k,:),X(tmp,:)));
      M(k,:) = X(tmp(idx),:);
    end
    l = km_le(X,M);
  end
  MM{i+1} = M; LL{i+1} = l;
end


% l = km_le(X,M) Labels given modes
%
% In:
%   X: data set.
%   M: current centroids.
% Out:
%   l: labels.

function l = km_le(X,M)

[~,l] = min(sqdist(X,M),[],2);	% Assignment of data points to modes


% M = km_m(X,sigma,l,M,I,tol[,deadM]) Modes given labels
%
% This finds the mode of each kde, which can be done with mean-shift,
% initialized from the previous iteration's modes.
%
% In:
%   X, l, M, I, tol: see above.
%   sigma: scalar containing the bandwidth value.
%   deadM: what to do with a dead mode (which has no associated data
%      points), one of 'rnd' (assign to it a random data point) or
%      'leave' (leave as is). Default: 'rnd'.
% Out:
%   M: new modes.

function M = km_m(X,sigma,l,M,I,tol,deadM)

% Argument defaults
if ~exist('deadM','var') || isempty(deadM) deadM = 'rnd'; end;

for k=1:size(M,1)
  tmp = find(l==k); ltmp = length(tmp);
  if ltmp > 0
    x = M(k,:);
    % Mean-shift iterations
    for i=1:I
      oldx = x;
      W = exp(-sqdist(x,X(tmp,:))/(2*sigma^2));
      x = (W*X(tmp,:))/sum(W);
      if max(abs(x-oldx) - tol*max(abs(oldx))) < 0 break; end
    end
    M(k,:) = x;
  elseif strcmp(deadM,'rnd')
    % Dead mode: assign to it one data point at random
    M(k,:) = X(floor(rand*size(X,1)+1),:);
    % Otherwise:
    % Leave dead mean as is (works badly because often modes end up dead)
  end
end


% M = km_init(X,K,C) Initialize modes
%
% In:
%   X, K, C: as above.
% Out:
%   M: as above.

function M = km_init(X,K,C)

[N,D] = size(X);
if ischar(C)
  %  rand('state',sum(100*clock));
  switch C
   case 'kmeans'
    % run kmeans multiple times and pick the best local optimum.
    errfunc = Inf;
    for i=1:20
      [tmp_M,~,tmp_errfunc] = kmeans(X,K);
      if tmp_errfunc(end) < errfunc errfunc = tmp_errfunc(end); M = tmp_M; end
    end
   case 'rndmeans'
    % M = uniform random points in the range of X
    m = min(X,[],1); mm = max(X,[],1);
    M = bsxfun(@plus,m,bsxfun(@times,mm-m,rand(K,D)));
   case 'rndsubset'
    % M = random subset of data points (works badly because often modes
    % compete inside clusters)
    tmp = randperm(N); M = X(tmp(1:K),:);
  end
else
  % M = user-provided
  M = C;
end

% s = bandwidths(X[,k,n]) List of bandwidth values for the homotopy
%
% Returns a geometric sequence of n values from s1 downto sn where:
% - s1 = 1/5 of the diameter of X
% - sn = average distance to the kth nearest neighbor
% This provides a reasonable coverage from large bandwidths (similar to
% k-means) to small ones. Other rules to estimate a good target bandwidth sn
% may be possible.
%
% In:
%   X: NxD dataset.
%   k: nearest neighbor order. Default: 10.
%   n: number of bandwidths in the sequence. Default: 20.
% Out:
%   s: list of bandwidths, in decreasing order.

function s = bandwidths(X,k,n)

% ---------- Argument defaults ----------
if ~exist('k','var') || isempty(k) k = 10; end;
if ~exist('n','var') || isempty(n) n = 20; end;
% ---------- End of "argument defaults" ----------

DD = sqdist(X);
s1 = sqrt(max(DD(:)))/5;	% diameter(X)/5
DD = sort(DD,2);
sn = mean(sqrt(DD(:,k+1)));	% average distance to kth nearest neighbor
s = 10.^linspace(log10(s1),log10(sn),n);	% Bandwidth values

